from __future__ import annotations

from dataclasses import dataclass
from typing import Literal


@dataclass
class ModelSettings:
    """Settings to use when calling an LLM.

    This class holds optional model configuration parameters (e.g. temperature,
    top_p, penalties, truncation, etc.).
    """

    temperature: float | None = None
    top_p: float | None = None
    frequency_penalty: float | None = None
    presence_penalty: float | None = None
    tool_choice: Literal["auto", "required", "none"] | str | None = None
    parallel_tool_calls: bool | None = False
    truncation: Literal["auto", "disabled"] | None = None

    def resolve(self, override: ModelSettings | None) -> ModelSettings:
        """Produce a new ModelSettings by overlaying any non-None values from the
        override on top of this instance."""
        if override is None:
            return self
        return ModelSettings(
            temperature=override.temperature or self.temperature,
            top_p=override.top_p or self.top_p,
            frequency_penalty=override.frequency_penalty or self.frequency_penalty,
            presence_penalty=override.presence_penalty or self.presence_penalty,
            tool_choice=override.tool_choice or self.tool_choice,
            parallel_tool_calls=override.parallel_tool_calls or self.parallel_tool_calls,
            truncation=override.truncation or self.truncation,
        )



from __future__ import annotations

import asyncio
from typing import Any

import pytest

from agents.tracing import (
    Span,
    Trace,
    agent_span,
    custom_span,
    function_span,
    generation_span,
    handoff_span,
    trace,
)
from agents.tracing.spans import SpanError

from .testing_processor import fetch_events, fetch_ordered_spans, fetch_traces

### HELPERS


def standard_span_checks(
    span: Span[Any], trace_id: str, parent_id: str | None, span_type: str
) -> None:
    assert span.span_id is not None
    assert span.trace_id == trace_id
    assert span.parent_id == parent_id
    assert span.started_at is not None
    assert span.ended_at is not None
    assert span.span_data.type == span_type


def standard_trace_checks(trace: Trace, name_check: str | None = None) -> None:
    assert trace.trace_id is not None

    if name_check:
        assert trace.name == name_check


### TESTS


def simple_tracing():
    x = trace("test")
    x.start()

    span_1 = agent_span(name="agent_1", parent=x)
    span_1.start()
    span_1.finish()

    span_2 = custom_span(name="custom_1", span_id="span_2", parent=x)
    span_2.start()

    span_3 = custom_span(name="custom_2", span_id="span_3", parent=span_2)
    span_3.start()
    span_3.finish()

    span_2.finish()

    x.finish()


def test_simple_tracing() -> None:
    simple_tracing()

    spans, traces = fetch_ordered_spans(), fetch_traces()
    assert len(spans) == 3
    assert len(traces) == 1

    trace = traces[0]
    standard_trace_checks(trace, name_check="test")
    trace_id = trace.trace_id

    first_span = spans[0]
    standard_span_checks(first_span, trace_id=trace_id, parent_id=None, span_type="agent")
    assert first_span.span_data.name == "agent_1"

    second_span = spans[1]
    standard_span_checks(second_span, trace_id=trace_id, parent_id=None, span_type="custom")
    assert second_span.span_id == "span_2"
    assert second_span.span_data.name == "custom_1"

    third_span = spans[2]
    standard_span_checks(
        third_span, trace_id=trace_id, parent_id=second_span.span_id, span_type="custom"
    )
    assert third_span.span_id == "span_3"
    assert third_span.span_data.name == "custom_2"


def ctxmanager_spans():
    with trace(workflow_name="test", trace_id="123", group_id="456"):
        with custom_span(name="custom_1", span_id="span_1"):
            with custom_span(name="custom_2", span_id="span_1_inner"):
                pass

        with custom_span(name="custom_2", span_id="span_2"):
            pass


def test_ctxmanager_spans() -> None:
    ctxmanager_spans()

    spans, traces = fetch_ordered_spans(), fetch_traces()
    assert len(spans) == 3
    assert len(traces) == 1

    trace = traces[0]
    standard_trace_checks(trace, name_check="test")
    trace_id = trace.trace_id

    first_span = spans[0]
    standard_span_checks(first_span, trace_id=trace_id, parent_id=None, span_type="custom")
    assert first_span.span_id == "span_1"

    first_inner_span = spans[1]
    standard_span_checks(
        first_inner_span, trace_id=trace_id, parent_id=first_span.span_id, span_type="custom"
    )
    assert first_inner_span.span_id == "span_1_inner"

    second_span = spans[2]
    standard_span_checks(second_span, trace_id=trace_id, parent_id=None, span_type="custom")
    assert second_span.span_id == "span_2"


async def run_subtask(span_id: str | None = None) -> None:
    with generation_span(span_id=span_id):
        await asyncio.sleep(0.01)


async def simple_async_tracing():
    with trace(workflow_name="test", trace_id="123", group_id="456"):
        await run_subtask(span_id="span_1")
        await run_subtask(span_id="span_2")


@pytest.mark.asyncio
async def test_async_tracing() -> None:
    await simple_async_tracing()

    spans, traces = fetch_ordered_spans(), fetch_traces()
    assert len(spans) == 2
    assert len(traces) == 1

    trace = traces[0]
    standard_trace_checks(trace, name_check="test")
    trace_id = trace.trace_id

    # We don't care about ordering here, just that they're there
    for s in spans:
        standard_span_checks(s, trace_id=trace_id, parent_id=None, span_type="generation")

    ids = [span.span_id for span in spans]
    assert "span_1" in ids
    assert "span_2" in ids


async def run_tasks_parallel(span_ids: list[str]) -> None:
    await asyncio.gather(
        *[run_subtask(span_id=span_id) for span_id in span_ids],
    )


async def run_tasks_as_children(first_span_id: str, second_span_id: str) -> None:
    with generation_span(span_id=first_span_id):
        await run_subtask(span_id=second_span_id)


async def complex_async_tracing():
    with trace(workflow_name="test", trace_id="123", group_id="456"):
        await asyncio.sleep(0.01)
        await asyncio.gather(
            run_tasks_parallel(["span_1", "span_2"]),
            run_tasks_parallel(["span_3", "span_4"]),
        )
        await asyncio.sleep(0.01)
        await asyncio.gather(
            run_tasks_as_children("span_5", "span_6"),
            run_tasks_as_children("span_7", "span_8"),
        )


@pytest.mark.asyncio
async def test_complex_async_tracing() -> None:
    await complex_async_tracing()

    spans, traces = fetch_ordered_spans(), fetch_traces()
    assert len(spans) == 8
    assert len(traces) == 1

    trace = traces[0]
    standard_trace_checks(trace, name_check="test")
    trace_id = trace.trace_id

    # First ensure 1,2,3,4 exist and are in parallel with the trace as parent
    for span_id in ["span_1", "span_2", "span_3", "span_4"]:
        span = next((s for s in spans if s.span_id == span_id), None)
        assert span is not None
        standard_span_checks(span, trace_id=trace_id, parent_id=None, span_type="generation")

    # Ensure 5 and 7 exist and have the trace as parent
    for span_id in ["span_5", "span_7"]:
        span = next((s for s in spans if s.span_id == span_id), None)
        assert span is not None
        standard_span_checks(span, trace_id=trace_id, parent_id=None, span_type="generation")

    # Ensure 6 and 8 exist and have 5 and 7 as parents
    six = next((s for s in spans if s.span_id == "span_6"), None)
    assert six is not None
    standard_span_checks(six, trace_id=trace_id, parent_id="span_5", span_type="generation")
    eight = next((s for s in spans if s.span_id == "span_8"), None)
    assert eight is not None
    standard_span_checks(eight, trace_id=trace_id, parent_id="span_7", span_type="generation")


def spans_with_setters():
    with trace(workflow_name="test", trace_id="123", group_id="456"):
        with agent_span(name="agent_1") as span_a:
            span_a.span_data.name = "agent_2"

            with function_span(name="function_1") as span_b:
                span_b.span_data.input = "i"
                span_b.span_data.output = "o"

            with generation_span() as span_c:
                span_c.span_data.input = [{"foo": "bar"}]

            with handoff_span(from_agent="agent_1", to_agent="agent_2"):
                pass


def test_spans_with_setters() -> None:
    spans_with_setters()

    spans, traces = fetch_ordered_spans(), fetch_traces()
    assert len(spans) == 4
    assert len(traces) == 1

    trace = traces[0]
    standard_trace_checks(trace, name_check="test")
    trace_id = trace.trace_id

    # Check the spans
    first_span = spans[0]
    standard_span_checks(first_span, trace_id=trace_id, parent_id=None, span_type="agent")
    assert first_span.span_data.name == "agent_2"

    second_span = spans[1]
    standard_span_checks(
        second_span, trace_id=trace_id, parent_id=first_span.span_id, span_type="function"
    )
    assert second_span.span_data.input == "i"
    assert second_span.span_data.output == "o"

    third_span = spans[2]
    standard_span_checks(
        third_span, trace_id=trace_id, parent_id=first_span.span_id, span_type="generation"
    )

    fourth_span = spans[3]
    standard_span_checks(
        fourth_span, trace_id=trace_id, parent_id=first_span.span_id, span_type="handoff"
    )


def disabled_tracing():
    with trace(workflow_name="test", trace_id="123", group_id="456", disabled=True):
        with agent_span(name="agent_1"):
            with function_span(name="function_1"):
                pass


def test_disabled_tracing():
    disabled_tracing()

    spans, traces = fetch_ordered_spans(), fetch_traces()
    assert len(spans) == 0
    assert len(traces) == 0


def enabled_trace_disabled_span():
    with trace(workflow_name="test", trace_id="123"):
        with agent_span(name="agent_1"):
            with function_span(name="function_1", disabled=True):
                with generation_span():
                    pass


def test_enabled_trace_disabled_span():
    enabled_trace_disabled_span()

    spans, traces = fetch_ordered_spans(), fetch_traces()
    assert len(spans) == 1  # Only the agent span is recorded
    assert len(traces) == 1  # The trace is recorded

    trace = traces[0]
    standard_trace_checks(trace, name_check="test")
    trace_id = trace.trace_id

    first_span = spans[0]
    standard_span_checks(first_span, trace_id=trace_id, parent_id=None, span_type="agent")
    assert first_span.span_data.name == "agent_1"


def test_start_and_end_called_manual():
    simple_tracing()

    events = fetch_events()

    assert events == [
        "trace_start",
        "span_start",  # span_1
        "span_end",  # span_1
        "span_start",  # span_2
        "span_start",  # span_3
        "span_end",  # span_3
        "span_end",  # span_2
        "trace_end",
    ]


def test_start_and_end_called_ctxmanager():
    with trace(workflow_name="test", trace_id="123", group_id="456"):
        with custom_span(name="custom_1", span_id="span_1"):
            with custom_span(name="custom_2", span_id="span_1_inner"):
                pass

        with custom_span(name="custom_2", span_id="span_2"):
            pass

    events = fetch_events()

    assert events == [
        "trace_start",
        "span_start",  # span_1
        "span_start",  # span_1_inner
        "span_end",  # span_1_inner
        "span_end",  # span_1
        "span_start",  # span_2
        "span_end",  # span_2
        "trace_end",
    ]


@pytest.mark.asyncio
async def test_start_and_end_called_async_ctxmanager():
    await simple_async_tracing()

    events = fetch_events()

    assert events == [
        "trace_start",
        "span_start",  # span_1
        "span_end",  # span_1
        "span_start",  # span_2
        "span_end",  # span_2
        "trace_end",
    ]


async def test_noop_span_doesnt_record():
    with trace(workflow_name="test", disabled=True) as t:
        with custom_span(name="span_1") as span:
            span.set_error(SpanError(message="test", data={}))

    spans, traces = fetch_ordered_spans(), fetch_traces()
    assert len(spans) == 0
    assert len(traces) == 0

    assert t.export() is None
    assert span.export() is None
    assert span.started_at is None
    assert span.ended_at is None
    assert span.error is None


async def test_multiple_span_start_finish_doesnt_crash():
    with trace(workflow_name="test", trace_id="123", group_id="456"):
        with custom_span(name="span_1") as span:
            span.start()

        span.finish()


async def test_noop_parent_is_noop_child():
    tr = trace(workflow_name="test", disabled=True)

    span = custom_span(name="span_1", parent=tr)
    span.start()
    span.finish()

    assert span.export() is None

    span_2 = custom_span(name="span_2", parent=span)
    span_2.start()
    span_2.finish()

    assert span_2.export() is None



from __future__ import annotations

import asyncio
import json
from typing import Any

import pytest
from typing_extensions import TypedDict

from agents import (
    Agent,
    AgentSpanData,
    FunctionSpanData,
    GenerationSpanData,
    GuardrailFunctionOutput,
    InputGuardrail,
    InputGuardrailTripwireTriggered,
    MaxTurnsExceeded,
    ModelBehaviorError,
    OutputGuardrail,
    OutputGuardrailTripwireTriggered,
    RunContextWrapper,
    Runner,
    TResponseInputItem,
)

from .fake_model import FakeModel
from .test_responses import (
    get_final_output_message,
    get_function_tool,
    get_function_tool_call,
    get_handoff_tool_call,
    get_text_message,
)
from .testing_processor import fetch_ordered_spans, fetch_traces


@pytest.mark.asyncio
async def test_single_turn_model_error():
    model = FakeModel(tracing_enabled=True)
    model.set_next_output(ValueError("test error"))

    agent = Agent(
        name="test_agent",
        model=model,
    )
    with pytest.raises(ValueError):
        result = Runner.run_streamed(agent, input="first_test")
        async for _ in result.stream_events():
            pass

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 2, f"should have agent and generation spans, got {len(spans)}"

    generation_span = spans[1]
    assert isinstance(generation_span.span_data, GenerationSpanData)
    assert generation_span.error, "should have error"


@pytest.mark.asyncio
async def test_multi_turn_no_handoffs():
    model = FakeModel(tracing_enabled=True)

    agent = Agent(
        name="test_agent",
        model=model,
        tools=[get_function_tool("foo", "tool_result")],
    )

    model.add_multiple_turn_outputs(
        [
            # First turn: a message and tool call
            [get_text_message("a_message"), get_function_tool_call("foo", json.dumps({"a": "b"}))],
            # Second turn: error
            ValueError("test error"),
            # Third turn: text message
            [get_text_message("done")],
        ]
    )

    with pytest.raises(ValueError):
        result = Runner.run_streamed(agent, input="first_test")
        async for _ in result.stream_events():
            pass

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 4, (
        f"should have agent, generation, tool, generation, got {len(spans)} with data: "
        f"{[x.span_data for x in spans]}"
    )

    last_generation_span = [x for x in spans if isinstance(x.span_data, GenerationSpanData)][-1]
    assert last_generation_span.error, "should have error"


@pytest.mark.asyncio
async def test_tool_call_error():
    model = FakeModel(tracing_enabled=True)

    agent = Agent(
        name="test_agent",
        model=model,
        tools=[get_function_tool("foo", "tool_result", hide_errors=True)],
    )

    model.set_next_output(
        [get_text_message("a_message"), get_function_tool_call("foo", "bad_json")],
    )

    with pytest.raises(ModelBehaviorError):
        result = Runner.run_streamed(agent, input="first_test")
        async for _ in result.stream_events():
            pass

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 3, (
        f"should have agent, generation, tool spans, got {len(spans)} with data: "
        f"{[x.span_data for x in spans]}"
    )

    function_span = [x for x in spans if isinstance(x.span_data, FunctionSpanData)][0]
    assert function_span.error, "should have error"


@pytest.mark.asyncio
async def test_multiple_handoff_doesnt_error():
    model = FakeModel(tracing_enabled=True)

    agent_1 = Agent(
        name="test",
        model=model,
    )
    agent_2 = Agent(
        name="test",
        model=model,
    )
    agent_3 = Agent(
        name="test",
        model=model,
        handoffs=[agent_1, agent_2],
        tools=[get_function_tool("some_function", "result")],
    )

    model.add_multiple_turn_outputs(
        [
            # First turn: a tool call
            [get_function_tool_call("some_function", json.dumps({"a": "b"}))],
            # Second turn: a message and 2 handoff
            [
                get_text_message("a_message"),
                get_handoff_tool_call(agent_1),
                get_handoff_tool_call(agent_2),
            ],
            # Third turn: text message
            [get_text_message("done")],
        ]
    )
    result = Runner.run_streamed(agent_3, input="user_message")
    async for _ in result.stream_events():
        pass

    assert result.last_agent == agent_1, "should have picked first handoff"

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 7, (
        f"should have 2 agent, 1 function, 3 generation, 1 handoff, got {len(spans)} with data: "
        f"{[x.span_data for x in spans]}"
    )


class Foo(TypedDict):
    bar: str


@pytest.mark.asyncio
async def test_multiple_final_output_no_error():
    model = FakeModel(tracing_enabled=True)

    agent_1 = Agent(
        name="test",
        model=model,
        output_type=Foo,
    )

    model.set_next_output(
        [
            get_final_output_message(json.dumps(Foo(bar="baz"))),
            get_final_output_message(json.dumps(Foo(bar="abc"))),
        ]
    )

    result = Runner.run_streamed(agent_1, input="user_message")
    async for _ in result.stream_events():
        pass

    assert isinstance(result.final_output, dict)
    assert result.final_output["bar"] == "abc"

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 2, (
        f"should have 1 agent, 1 generation, got {len(spans)} with data: "
        f"{[x.span_data for x in spans]}"
    )


@pytest.mark.asyncio
async def test_handoffs_lead_to_correct_agent_spans():
    model = FakeModel(tracing_enabled=True)

    agent_1 = Agent(
        name="test_agent_1",
        model=model,
        tools=[get_function_tool("some_function", "result")],
    )
    agent_2 = Agent(
        name="test_agent_2",
        model=model,
        handoffs=[agent_1],
        tools=[get_function_tool("some_function", "result")],
    )
    agent_3 = Agent(
        name="test_agent_3",
        model=model,
        handoffs=[agent_1, agent_2],
        tools=[get_function_tool("some_function", "result")],
    )

    agent_1.handoffs.append(agent_3)

    model.add_multiple_turn_outputs(
        [
            # First turn: a tool call
            [get_function_tool_call("some_function", json.dumps({"a": "b"}))],
            # Second turn: a message and 2 handoff
            [
                get_text_message("a_message"),
                get_handoff_tool_call(agent_1),
                get_handoff_tool_call(agent_2),
            ],
            # Third turn: tool call
            [get_function_tool_call("some_function", json.dumps({"a": "b"}))],
            # Fourth turn: handoff
            [get_handoff_tool_call(agent_3)],
            # Fifth turn: text message
            [get_text_message("done")],
        ]
    )
    result = Runner.run_streamed(agent_3, input="user_message")
    async for _ in result.stream_events():
        pass

    assert result.last_agent == agent_3, (
        f"should have ended on the third agent, got {result.last_agent.name}"
    )

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 12, (
        f"should have 3 agents, 2 function, 5 generation, 2 handoff, got {len(spans)} with data: "
        f"{[x.span_data for x in spans]}"
    )


@pytest.mark.asyncio
async def test_max_turns_exceeded():
    model = FakeModel(tracing_enabled=True)

    agent = Agent(
        name="test",
        model=model,
        output_type=Foo,
        tools=[get_function_tool("foo", "result")],
    )

    model.add_multiple_turn_outputs(
        [
            [get_function_tool_call("foo")],
            [get_function_tool_call("foo")],
            [get_function_tool_call("foo")],
            [get_function_tool_call("foo")],
            [get_function_tool_call("foo")],
        ]
    )

    with pytest.raises(MaxTurnsExceeded):
        result = Runner.run_streamed(agent, input="user_message", max_turns=2)
        async for _ in result.stream_events():
            pass

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 5, (
        f"should have 1 agent, 2 generations, 2 function calls, got "
        f"{len(spans)} with data: {[x.span_data for x in spans]}"
    )

    agent_span = [x for x in spans if isinstance(x.span_data, AgentSpanData)][-1]
    assert agent_span.error, "last agent should have error"


def input_guardrail_function(
    context: RunContextWrapper[Any], agent: Agent[Any], input: str | list[TResponseInputItem]
) -> GuardrailFunctionOutput:
    return GuardrailFunctionOutput(
        output_info=None,
        tripwire_triggered=True,
    )


@pytest.mark.asyncio
async def test_input_guardrail_error():
    model = FakeModel()

    agent = Agent(
        name="test",
        model=model,
        input_guardrails=[InputGuardrail(guardrail_function=input_guardrail_function)],
    )
    model.set_next_output([get_text_message("some_message")])

    with pytest.raises(InputGuardrailTripwireTriggered):
        result = Runner.run_streamed(agent, input="user_message")
        async for _ in result.stream_events():
            pass

    await asyncio.sleep(1)

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 2, (
        f"should have 1 agent, 1 guardrail, got {len(spans)} with data: "
        f"{[x.span_data for x in spans]}"
    )

    agent_span = [x for x in spans if isinstance(x.span_data, AgentSpanData)][-1]
    assert agent_span.error, "last agent should have error"


def output_guardrail_function(
    context: RunContextWrapper[Any], agent: Agent[Any], agent_output: Any
) -> GuardrailFunctionOutput:
    return GuardrailFunctionOutput(
        output_info=None,
        tripwire_triggered=True,
    )


@pytest.mark.asyncio
async def test_output_guardrail_error():
    model = FakeModel()

    agent = Agent(
        name="test",
        model=model,
        output_guardrails=[OutputGuardrail(guardrail_function=output_guardrail_function)],
    )
    model.set_next_output([get_text_message("some_message")])

    with pytest.raises(OutputGuardrailTripwireTriggered):
        result = Runner.run_streamed(agent, input="user_message")
        async for _ in result.stream_events():
            pass

    await asyncio.sleep(1)

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 2, (
        f"should have 1 agent, 1 guardrail, got {len(spans)} with data: "
        f"{[x.span_data for x in spans]}"
    )

    agent_span = [x for x in spans if isinstance(x.span_data, AgentSpanData)][-1]
    assert agent_span.error, "last agent should have error"



from __future__ import annotations

import threading
from typing import Any, Literal

from agents.tracing import Span, Trace, TracingProcessor

TestSpanProcessorEvent = Literal["trace_start", "trace_end", "span_start", "span_end"]


class SpanProcessorForTests(TracingProcessor):
    """
    A simple processor that stores finished spans in memory.
    This is thread-safe and suitable for tests or basic usage.
    """

    def __init__(self) -> None:
        self._lock = threading.Lock()
        # Dictionary of trace_id -> list of spans
        self._spans: list[Span[Any]] = []
        self._traces: list[Trace] = []
        self._events: list[TestSpanProcessorEvent] = []

    def on_trace_start(self, trace: Trace) -> None:
        with self._lock:
            self._traces.append(trace)
            self._events.append("trace_start")

    def on_trace_end(self, trace: Trace) -> None:
        with self._lock:
            # We don't append the trace here, we want to do that in on_trace_start
            self._events.append("trace_end")

    def on_span_start(self, span: Span[Any]) -> None:
        with self._lock:
            # Purposely not appending the span here, we want to do that in on_span_end
            self._events.append("span_start")

    def on_span_end(self, span: Span[Any]) -> None:
        with self._lock:
            self._events.append("span_end")
            self._spans.append(span)

    def get_ordered_spans(self, including_empty: bool = False) -> list[Span[Any]]:
        with self._lock:
            spans = [x for x in self._spans if including_empty or x.export()]
            return sorted(spans, key=lambda x: x.started_at or 0)

    def get_traces(self, including_empty: bool = False) -> list[Trace]:
        with self._lock:
            traces = [x for x in self._traces if including_empty or x.export()]
            return traces

    def clear(self) -> None:
        with self._lock:
            self._spans.clear()
            self._traces.clear()
            self._events.clear()

    def shutdown(self) -> None:
        pass

    def force_flush(self) -> None:
        pass


SPAN_PROCESSOR_TESTING = SpanProcessorForTests()


def fetch_ordered_spans() -> list[Span[Any]]:
    return SPAN_PROCESSOR_TESTING.get_ordered_spans()


def fetch_traces() -> list[Trace]:
    return SPAN_PROCESSOR_TESTING.get_traces()


def fetch_events() -> list[TestSpanProcessorEvent]:
    return SPAN_PROCESSOR_TESTING._events



from __future__ import annotations

import abc
import copy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar, Union

from openai.types.responses import (
    Response,
    ResponseComputerToolCall,
    ResponseFileSearchToolCall,
    ResponseFunctionToolCall,
    ResponseFunctionWebSearch,
    ResponseInputItemParam,
    ResponseOutputItem,
    ResponseOutputMessage,
    ResponseOutputRefusal,
    ResponseOutputText,
    ResponseStreamEvent,
)
from openai.types.responses.response_input_item_param import ComputerCallOutput, FunctionCallOutput
from openai.types.responses.response_reasoning_item import ResponseReasoningItem
from pydantic import BaseModel
from typing_extensions import TypeAlias

from .exceptions import AgentsException, ModelBehaviorError
from .usage import Usage

if TYPE_CHECKING:
    from .agent import Agent

TResponse = Response
"""A type alias for the Response type from the OpenAI SDK."""

TResponseInputItem = ResponseInputItemParam
"""A type alias for the ResponseInputItemParam type from the OpenAI SDK."""

TResponseOutputItem = ResponseOutputItem
"""A type alias for the ResponseOutputItem type from the OpenAI SDK."""

TResponseStreamEvent = ResponseStreamEvent
"""A type alias for the ResponseStreamEvent type from the OpenAI SDK."""

T = TypeVar("T", bound=Union[TResponseOutputItem, TResponseInputItem])


@dataclass
class RunItemBase(Generic[T], abc.ABC):
    agent: Agent[Any]
    """The agent whose run caused this item to be generated."""

    raw_item: T
    """The raw Responses item from the run. This will always be a either an output item (i.e.
    `openai.types.responses.ResponseOutputItem` or an input item
    (i.e. `openai.types.responses.ResponseInputItemParam`).
    """

    def to_input_item(self) -> TResponseInputItem:
        """Converts this item into an input item suitable for passing to the model."""
        if isinstance(self.raw_item, dict):
            # We know that input items are dicts, so we can ignore the type error
            return self.raw_item  # type: ignore
        elif isinstance(self.raw_item, BaseModel):
            # All output items are Pydantic models that can be converted to input items.
            return self.raw_item.model_dump(exclude_unset=True)  # type: ignore
        else:
            raise AgentsException(f"Unexpected raw item type: {type(self.raw_item)}")


@dataclass
class MessageOutputItem(RunItemBase[ResponseOutputMessage]):
    """Represents a message from the LLM."""

    raw_item: ResponseOutputMessage
    """The raw response output message."""

    type: Literal["message_output_item"] = "message_output_item"


@dataclass
class HandoffCallItem(RunItemBase[ResponseFunctionToolCall]):
    """Represents a tool call for a handoff from one agent to another."""

    raw_item: ResponseFunctionToolCall
    """The raw response function tool call that represents the handoff."""

    type: Literal["handoff_call_item"] = "handoff_call_item"


@dataclass
class HandoffOutputItem(RunItemBase[TResponseInputItem]):
    """Represents the output of a handoff."""

    raw_item: TResponseInputItem
    """The raw input item that represents the handoff taking place."""

    source_agent: Agent[Any]
    """The agent that made the handoff."""

    target_agent: Agent[Any]
    """The agent that is being handed off to."""

    type: Literal["handoff_output_item"] = "handoff_output_item"


ToolCallItemTypes: TypeAlias = Union[
    ResponseFunctionToolCall,
    ResponseComputerToolCall,
    ResponseFileSearchToolCall,
    ResponseFunctionWebSearch,
]
"""A type that represents a tool call item."""


@dataclass
class ToolCallItem(RunItemBase[ToolCallItemTypes]):
    """Represents a tool call e.g. a function call or computer action call."""

    raw_item: ToolCallItemTypes
    """The raw tool call item."""

    type: Literal["tool_call_item"] = "tool_call_item"


@dataclass
class ToolCallOutputItem(RunItemBase[Union[FunctionCallOutput, ComputerCallOutput]]):
    """Represents the output of a tool call."""

    raw_item: FunctionCallOutput | ComputerCallOutput
    """The raw item from the model."""

    output: str
    """The output of the tool call."""

    type: Literal["tool_call_output_item"] = "tool_call_output_item"


@dataclass
class ReasoningItem(RunItemBase[ResponseReasoningItem]):
    """Represents a reasoning item."""

    raw_item: ResponseReasoningItem
    """The raw reasoning item."""

    type: Literal["reasoning_item"] = "reasoning_item"


RunItem: TypeAlias = Union[
    MessageOutputItem,
    HandoffCallItem,
    HandoffOutputItem,
    ToolCallItem,
    ToolCallOutputItem,
    ReasoningItem,
]
"""An item generated by an agent."""


@dataclass
class ModelResponse:
    output: list[TResponseOutputItem]
    """A list of outputs (messages, tool calls, etc) generated by the model"""

    usage: Usage
    """The usage information for the response."""

    referenceable_id: str | None
    """An ID for the response which can be used to refer to the response in subsequent calls to the
    model. Not supported by all model providers.
    """

    def to_input_items(self) -> list[TResponseInputItem]:
        """Convert the output into a list of input items suitable for passing to the model."""
        # We happen to know that the shape of the Pydantic output items are the same as the
        # equivalent TypedDict input items, so we can just convert each one.
        # This is also tested via unit tests.
        return [it.model_dump(exclude_unset=True) for it in self.output]  # type: ignore


class ItemHelpers:
    @classmethod
    def extract_last_content(cls, message: TResponseOutputItem) -> str:
        """Extracts the last text content or refusal from a message."""
        if not isinstance(message, ResponseOutputMessage):
            return ""

        last_content = message.content[-1]
        if isinstance(last_content, ResponseOutputText):
            return last_content.text
        elif isinstance(last_content, ResponseOutputRefusal):
            return last_content.refusal
        else:
            raise ModelBehaviorError(f"Unexpected content type: {type(last_content)}")

    @classmethod
    def extract_last_text(cls, message: TResponseOutputItem) -> str | None:
        """Extracts the last text content from a message, if any. Ignores refusals."""
        if isinstance(message, ResponseOutputMessage):
            last_content = message.content[-1]
            if isinstance(last_content, ResponseOutputText):
                return last_content.text

        return None

    @classmethod
    def input_to_new_input_list(
        cls, input: str | list[TResponseInputItem]
    ) -> list[TResponseInputItem]:
        """Converts a string or list of input items into a list of input items."""
        if isinstance(input, str):
            return [
                {
                    "content": input,
                    "role": "user",
                }
            ]
        return copy.deepcopy(input)

    @classmethod
    def text_message_outputs(cls, items: list[RunItem]) -> str:
        """Concatenates all the text content from a list of message output items."""
        text = ""
        for item in items:
            if isinstance(item, MessageOutputItem):
                text += cls.text_message_output(item)
        return text

    @classmethod
    def text_message_output(cls, message: MessageOutputItem) -> str:
        """Extracts all the text content from a single message output item."""
        text = ""
        for item in message.raw_item.content:
            if isinstance(item, ResponseOutputText):
                text += item.text
        return text

    @classmethod
    def tool_call_output_item(
        cls, tool_call: ResponseFunctionToolCall, output: str
    ) -> FunctionCallOutput:
        """Creates a tool call output item from a tool call and its output."""
        return {
            "call_id": tool_call.call_id,
            "output": output,
            "type": "function_call_output",
        }



from __future__ import annotations

import asyncio

import pytest

from agents import Agent, RunConfig, Runner, trace

from .fake_model import FakeModel
from .test_responses import get_text_message
from .testing_processor import fetch_ordered_spans, fetch_traces


@pytest.mark.asyncio
async def test_single_run_is_single_trace():
    agent = Agent(
        name="test_agent",
        model=FakeModel(
            initial_output=[get_text_message("first_test")],
        ),
    )

    await Runner.run(agent, input="first_test")

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 1, (
        f"Got {len(spans)}, but expected 1: the agent span. data:"
        f"{[span.span_data for span in spans]}"
    )


@pytest.mark.asyncio
async def test_multiple_runs_are_multiple_traces():
    model = FakeModel()
    model.add_multiple_turn_outputs(
        [
            [get_text_message("first_test")],
            [get_text_message("second_test")],
        ]
    )
    agent = Agent(
        name="test_agent_1",
        model=model,
    )

    await Runner.run(agent, input="first_test")
    await Runner.run(agent, input="second_test")

    traces = fetch_traces()
    assert len(traces) == 2, f"Expected 2 traces, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 2, f"Got {len(spans)}, but expected 2: agent span per run"


@pytest.mark.asyncio
async def test_wrapped_trace_is_single_trace():
    model = FakeModel()
    model.add_multiple_turn_outputs(
        [
            [get_text_message("first_test")],
            [get_text_message("second_test")],
            [get_text_message("third_test")],
        ]
    )
    with trace(workflow_name="test_workflow"):
        agent = Agent(
            name="test_agent_1",
            model=model,
        )

        await Runner.run(agent, input="first_test")
        await Runner.run(agent, input="second_test")
        await Runner.run(agent, input="third_test")

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 3, f"Got {len(spans)}, but expected 3: the agent span per run"


@pytest.mark.asyncio
async def test_parent_disabled_trace_disabled_agent_trace():
    with trace(workflow_name="test_workflow", disabled=True):
        agent = Agent(
            name="test_agent",
            model=FakeModel(
                initial_output=[get_text_message("first_test")],
            ),
        )

        await Runner.run(agent, input="first_test")

    traces = fetch_traces()
    assert len(traces) == 0, f"Expected 0 traces, got {len(traces)}"
    spans = fetch_ordered_spans()
    assert len(spans) == 0, (
        f"Expected no spans, got {len(spans)}, with {[x.span_data for x in spans]}"
    )


@pytest.mark.asyncio
async def test_manual_disabling_works():
    agent = Agent(
        name="test_agent",
        model=FakeModel(
            initial_output=[get_text_message("first_test")],
        ),
    )

    await Runner.run(agent, input="first_test", run_config=RunConfig(tracing_disabled=True))

    traces = fetch_traces()
    assert len(traces) == 0, f"Expected 0 traces, got {len(traces)}"
    spans = fetch_ordered_spans()
    assert len(spans) == 0, f"Got {len(spans)}, but expected no spans"


@pytest.mark.asyncio
async def test_trace_config_works():
    agent = Agent(
        name="test_agent",
        model=FakeModel(
            initial_output=[get_text_message("first_test")],
        ),
    )

    await Runner.run(
        agent,
        input="first_test",
        run_config=RunConfig(workflow_name="Foo bar", group_id="123", trace_id="456"),
    )

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"
    export = traces[0].export()
    assert export is not None, "Trace export should not be None"
    assert export["workflow_name"] == "Foo bar"
    assert export["group_id"] == "123"
    assert export["id"] == "456"


@pytest.mark.asyncio
async def test_not_starting_streaming_creates_trace():
    agent = Agent(
        name="test_agent",
        model=FakeModel(
            initial_output=[get_text_message("first_test")],
        ),
    )

    result = Runner.run_streamed(agent, input="first_test")

    # Purposely don't await the stream
    while True:
        if result.is_complete:
            break
        await asyncio.sleep(0.1)

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"

    spans = fetch_ordered_spans()
    assert len(spans) == 1, f"Got {len(spans)}, but expected 1: the agent span"

    # Await the stream to avoid warnings about it not being awaited
    async for _ in result.stream_events():
        pass


@pytest.mark.asyncio
async def test_streaming_single_run_is_single_trace():
    agent = Agent(
        name="test_agent",
        model=FakeModel(
            initial_output=[get_text_message("first_test")],
        ),
    )

    x = Runner.run_streamed(agent, input="first_test")
    async for _ in x.stream_events():
        pass

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"


@pytest.mark.asyncio
async def test_multiple_streamed_runs_are_multiple_traces():
    model = FakeModel()
    model.add_multiple_turn_outputs(
        [
            [get_text_message("first_test")],
            [get_text_message("second_test")],
        ]
    )
    agent = Agent(
        name="test_agent_1",
        model=model,
    )

    x = Runner.run_streamed(agent, input="first_test")
    async for _ in x.stream_events():
        pass

    x = Runner.run_streamed(agent, input="second_test")
    async for _ in x.stream_events():
        pass

    traces = fetch_traces()
    assert len(traces) == 2, f"Expected 2 traces, got {len(traces)}"


@pytest.mark.asyncio
async def test_wrapped_streaming_trace_is_single_trace():
    model = FakeModel()
    model.add_multiple_turn_outputs(
        [
            [get_text_message("first_test")],
            [get_text_message("second_test")],
            [get_text_message("third_test")],
        ]
    )
    with trace(workflow_name="test_workflow"):
        agent = Agent(
            name="test_agent_1",
            model=model,
        )

        x = Runner.run_streamed(agent, input="first_test")
        async for _ in x.stream_events():
            pass

        x = Runner.run_streamed(agent, input="second_test")
        async for _ in x.stream_events():
            pass

        x = Runner.run_streamed(agent, input="third_test")
        async for _ in x.stream_events():
            pass

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"


@pytest.mark.asyncio
async def test_wrapped_mixed_trace_is_single_trace():
    model = FakeModel()
    model.add_multiple_turn_outputs(
        [
            [get_text_message("first_test")],
            [get_text_message("second_test")],
            [get_text_message("third_test")],
        ]
    )
    with trace(workflow_name="test_workflow"):
        agent = Agent(
            name="test_agent_1",
            model=model,
        )

        x = Runner.run_streamed(agent, input="first_test")
        async for _ in x.stream_events():
            pass

        await Runner.run(agent, input="second_test")

        x = Runner.run_streamed(agent, input="third_test")
        async for _ in x.stream_events():
            pass

    traces = fetch_traces()
    assert len(traces) == 1, f"Expected 1 trace, got {len(traces)}"


@pytest.mark.asyncio
async def test_parent_disabled_trace_disables_streaming_agent_trace():
    model = FakeModel()
    model.add_multiple_turn_outputs(
        [
            [get_text_message("first_test")],
            [get_text_message("second_test")],
        ]
    )
    with trace(workflow_name="test_workflow", disabled=True):
        agent = Agent(
            name="test_agent",
            model=model,
        )

        x = Runner.run_streamed(agent, input="first_test")
        async for _ in x.stream_events():
            pass

    traces = fetch_traces()
    assert len(traces) == 0, f"Expected 0 traces, got {len(traces)}"


@pytest.mark.asyncio
async def test_manual_streaming_disabling_works():
    model = FakeModel()
    model.add_multiple_turn_outputs(
        [
            [get_text_message("first_test")],
            [get_text_message("second_test")],
        ]
    )
    agent = Agent(
        name="test_agent",
        model=model,
    )

    x = Runner.run_streamed(agent, input="first_test", run_config=RunConfig(tracing_disabled=True))
    async for _ in x.stream_events():
        pass

    traces = fetch_traces()
    assert len(traces) == 0, f"Expected 0 traces, got {len(traces)}"



from __future__ import annotations

import httpx
from openai import AsyncOpenAI, DefaultAsyncHttpxClient

from . import _openai_shared
from .interface import Model, ModelProvider
from .openai_chatcompletions import OpenAIChatCompletionsModel
from .openai_responses import OpenAIResponsesModel

DEFAULT_MODEL: str = "gpt-4o"


_http_client: httpx.AsyncClient | None = None


# If we create a new httpx client for each request, that would mean no sharing of connection pools,
# which would mean worse latency and resource usage. So, we share the client across requests.
def shared_http_client() -> httpx.AsyncClient:
    global _http_client
    if _http_client is None:
        _http_client = DefaultAsyncHttpxClient()
    return _http_client


class OpenAIProvider(ModelProvider):
    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
        openai_client: AsyncOpenAI | None = None,
        organization: str | None = None,
        project: str | None = None,
        use_responses: bool | None = None,
    ) -> None:
        if openai_client is not None:
            assert api_key is None and base_url is None, (
                "Don't provide api_key or base_url if you provide openai_client"
            )
            self._client = openai_client
        else:
            self._client = _openai_shared.get_default_openai_client() or AsyncOpenAI(
                api_key=api_key or _openai_shared.get_default_openai_key(),
                base_url=base_url,
                organization=organization,
                project=project,
                http_client=shared_http_client(),
            )

        self._is_openai_model = self._client.base_url.host.startswith("api.openai.com")
        if use_responses is not None:
            self._use_responses = use_responses
        else:
            self._use_responses = _openai_shared.get_use_responses_by_default()

    def get_model(self, model_name: str | None) -> Model:
        if model_name is None:
            model_name = DEFAULT_MODEL

        return (
            OpenAIResponsesModel(model=model_name, openai_client=self._client)
            if self._use_responses
            else OpenAIChatCompletionsModel(model=model_name, openai_client=self._client)
        )



"""
Example demonstrating how to use an agent with MCP servers for Slack integration.

This example shows how to:
1. Load MCP servers from the config file automatically
2. Create an agent that connects to Slack via MCP tools
3. Run the agent to search through Slack conversations

To use this example:
1. Ensure you have a slack MCP server configured in your mcp_agent.config.yaml
2. Run this example
"""

import asyncio
import os
import sys
import time
from typing import TYPE_CHECKING, Any

from openai.types.responses import ResponseTextDeltaEvent

if TYPE_CHECKING:
    from mcp_agent.config import MCPSettings

# Add the src directory to the path for imports to work
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../../..')))

from agents import Agent, Runner

# Enable logging for debugging
# enable_verbose_stdout_logging()


class AgentContext:
    """Context class for the agent that can hold MCP settings."""

    def __init__(self, mcp_config_path: str = None, mcp_config: "MCPSettings" = None):
        """
        Initialize the context.

        Args:
            mcp_config_path: Optional path to the mcp_agent.config.yaml file
            mcp_config: Optional MCPSettings object
        """
        self.mcp_config_path = mcp_config_path
        self.mcp_config = mcp_config


async def main():
    """Run the Slack integration example."""
    start = time.time()

    # Create a context object
    context = AgentContext()

    # Create an agent that specifies which MCP servers to use
    # Make sure these are defined in your mcp_agent.config.yaml file
    agent = Agent(
        name="Slack Finder",
        instructions="""You are an agent with access to the filesystem, 
        as well as the ability to look up Slack conversations. Your job is to identify
        the closest match to a user's request, make the appropriate tool calls,
        and return the results.""",
        # Local tools can be added here if needed
        tools=[],
        # Specify which MCP servers to use
        mcp_servers=["filesystem", "slack"],
    )

    # First example: Search for last message in general channel
    print("\n\n--- FIRST QUERY ---")
    print("Searching for the last message in the general channel...\n")

    result = Runner.run_streamed(
        agent,
        input="What was the last message in the general channel?",
        context=context,
    )

    # Stream the response
    async for event in result.stream_events():
        if event.type == "raw_response_event" and isinstance(event.data, ResponseTextDeltaEvent):
            print(event.data.delta, end="", flush=True)

    # Second example: Follow-up question demonstrating multi-turn capabilities
    print("\n\n--- FOLLOW-UP QUERY ---")
    print("Asking for a summary of the returned information...\n")

    result = Runner.run_streamed(
        agent,
        input=f"Summarize {result.final_output} for me and save it as convo.txt in the current directory.",
        context=context,
    )

    # Stream the response
    async for event in result.stream_events():
        if event.type == "raw_response_event" and isinstance(event.data, ResponseTextDeltaEvent):
            print(event.data.delta, end="", flush=True)

    # Calculate and display total runtime
    end = time.time()
    t = end - start
    print(f"\n\nTotal run time: {t:.2f}s")


if __name__ == "__main__":
    asyncio.run(main())



"""
Example demonstrating how to use an agent with MCP servers.

This example shows how to:
1. Load MCP servers from the config file automatically
2. Create an agent that specifies which MCP servers to use
3. Run the agent to dynamically load and use tools from the specified MCP servers

To use this example:
1. Create an mcp_agent.config.yaml file in this directory or a parent directory
2. Configure your MCP servers in that file
3. Run this example
"""

import asyncio
import os
import sys
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from mcp_agent.config import MCPSettings

# Add the src directory to the path for imports to work
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from agents import Agent, Runner, function_tool, enable_verbose_stdout_logging

enable_verbose_stdout_logging()


# Define a simple local tool to demonstrate combining local and MCP tools
@function_tool
def get_current_weather(location: str) -> str:
    """
    Get the current weather for a location.

    Args:
        location: The city and state, e.g. "San Francisco, CA"

    Returns:
        The current weather for the requested location
    """
    return f"The weather in {location} is currently sunny and 72 degrees Fahrenheit."

# Create a simple context class that can be extended with MCP server registry
class AgentContext:
    def __init__(self, mcp_config: "MCPSettings" = None, mcp_config_path: str = None):
        """
        Initialize the context.

        Args:
            mcp_config: Optional MCPSettings object containing the server configurations
                If unspecified, the MCP settings are loaded from the mcp_config_path
            mcp_config_path: Optional path to the mcp_agent.config.yaml file
                If both mcp_config and mcp_config_path are unspecified, 
                the default discovery process will look for the config file matching
                "mcp_agent.config.yaml" recursively up from the current working directory.
        """
        self.mcp_config_path = mcp_config_path
        self.mcp_config = mcp_config


async def main():
    # Specify a custom config path if needed, or set to None to use default discovery
    mcp_config_path = None  # Set to a file path if needed

    # Alternatively, define MCP config programmatically
    mcp_config = None
    # mcp_config = MCPSettings(
    #     servers={
    #         "fetch": MCPServerSettings(
    #             command="uvx",
    #             args=["mcp-server-fetch"],
    #         ),
    #         "filesystem": MCPServerSettings(
    #             command="npx",
    #             args=["-y", "@modelcontextprotocol/server-filesystem", "."],
    #         ),
    #     }
    # ),

    # Create a context object containing MCP settings
    context = AgentContext(mcp_config_path=mcp_config_path, mcp_config=mcp_config)

    # Create an agent with specific MCP servers you want to use
    # These must be defined in your mcp_agent.config.yaml file
    agent = Agent(
        name="MCP Assistant",
        instructions="""You are a helpful assistant with access to both local tools 
            and tools from MCP servers. Use these tools to help the user.""",
        tools=[get_current_weather],  # Local tools
        mcp_servers=["fetch", "filesystem"],  # Specify which MCP servers to use
                                              # These must be defined in your config
    )

    # Run the agent - tools from the specified MCP servers will be automatically loaded
    result = await Runner.run(
        starting_agent=agent,
        input="What's the weather like in Miami?",
        context=context,
    )

    # Print the agent's response
    print("\nAgent response:")
    print(result.final_output)

    result = await Runner.run(
        starting_agent=agent,
        input="Print the first paragraph of https://openai.github.io/openai-agents-python/",
        context=context,
    )

    # Print the agent's response
    print("\nAgent response:")
    print(result.final_output)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import re
from collections.abc import Awaitable
from typing import Any, Literal, Union

from pydantic import TypeAdapter, ValidationError
from typing_extensions import TypeVar

from .exceptions import ModelBehaviorError
from .logger import logger
from .tracing import Span, SpanError, get_current_span

T = TypeVar("T")

MaybeAwaitable = Union[Awaitable[T], T]


def transform_string_function_style(name: str) -> str:
    # Replace spaces with underscores
    name = name.replace(" ", "_")

    # Replace non-alphanumeric characters with underscores
    name = re.sub(r"[^a-zA-Z0-9]", "_", name)

    return name.lower()


def validate_json(json_str: str, type_adapter: TypeAdapter[T], partial: bool) -> T:
    partial_setting: bool | Literal["off", "on", "trailing-strings"] = (
        "trailing-strings" if partial else False
    )
    try:
        validated = type_adapter.validate_json(json_str, experimental_allow_partial=partial_setting)
        return validated
    except ValidationError as e:
        attach_error_to_current_span(
            SpanError(
                message="Invalid JSON provided",
                data={},
            )
        )
        raise ModelBehaviorError(
            f"Invalid JSON when parsing {json_str} for {type_adapter}; {e}"
        ) from e


def attach_error_to_span(span: Span[Any], error: SpanError) -> None:
    span.set_error(error)


def attach_error_to_current_span(error: SpanError) -> None:
    span = get_current_span()
    if span:
        attach_error_to_span(span, error)
    else:
        logger.warning(f"No span to add error {error} to")


async def noop_coroutine() -> None:
    pass



from __future__ import annotations

from typing import Any, Dict, List, Set

from mcp.types import EmbeddedResource, ImageContent, TextContent, Tool as MCPTool
from mcp_agent.config import MCPSettings, Settings, get_settings
from mcp_agent.context import Context
from mcp_agent.mcp.mcp_aggregator import MCPAggregator
from mcp_agent.mcp_server_registry import ServerRegistry

from . import _utils
from .logger import logger
from .run_context import RunContextWrapper, TContext
from .tool import FunctionTool, Tool

MCPContent = TextContent | ImageContent | EmbeddedResource

# Define a method to automatically load MCP server registry if using MCP servers
def load_mcp_server_registry(
    config: MCPSettings | None = None,
    config_path: str | None = None
) -> ServerRegistry:
    """
    Load MCP server registry from config object or config file path.

    Args:
        config: The MCPSettings object containing the server configurations.
            If unspecified, it will be loaded from the config_path.
        config_path: The file path to load the MCP server configurations from.
            if config is unspecified, this is required.
    """
    try:
        settings: Settings = None
        if config:
            # Use provided settings object
            logger.debug("Loading MCP server registry from provided MCPSettings object.")
            settings = Settings(mcp=config)
        else:
            # Load settings from config file
            logger.debug("Loading MCP server registry from config file: %s", config_path)
            settings = get_settings(config_path)

        # Create the ServerRegistry instance
        server_registry = ServerRegistry(config=settings)
        return server_registry
    except Exception as e:
        logger.error(
            "Error loading MCP server registry. config=%s, config_path=%s, Error: %s",
            config.model_dump_json() if config else "None",
            config_path,
            e
        )
        raise

def ensure_mcp_server_registry_in_context(
    run_context: RunContextWrapper[TContext],
    force: bool = False
) -> ServerRegistry:
    """
    Load the MCP server registry and attach it to the context object.
    If the server registry is already loaded, it will b

    Args:
        run_context: Run context wrapper which will have the server registry attached
        force: Whether to force reload the server registry
    """
    # Check if server registry is already loaded
    server_registry = getattr(run_context.context, 'server_registry', None)
    if not force and server_registry:
        logger.debug("MCP server registry already loaded in context. Skipping reload.")
        return server_registry

    # Load the server registry
    config = getattr(run_context.context, 'mcp_config', None)
    config_path = getattr(run_context.context, 'mcp_config_path', None)
    server_registry = load_mcp_server_registry(config=config, config_path=config_path)

    # Attach the server registry to the context
    run_context.context.mcp_server_registry = server_registry

def create_mcp_aggregator(
    run_context: RunContextWrapper[TContext],
    name: str,
    servers: list[str],
    server_registry: ServerRegistry | None = None,
    connection_persistence: bool = True
) -> MCPAggregator:
    """
    Create the MCP aggregator with the MCP servers from server registry.
    This doesn't initialize the aggregator. For initialization, use `initialize_mcp_aggregator`.

    Args:
        run_context: Run context wrapper
        name: Name of the agent using the aggregator
        servers: List of MCP server names
        server_registry: Server registry instance (if not provided, it will be retrieved from context)
        connection_persistence: Whether to keep the server connections alive, or restart per call
    """
    if not servers:
        raise RuntimeError("No MCP servers specified. No MCP aggregator created.")

    # Get or create the server registry from the context
    context: Context = None
    if server_registry:
        context = Context(server_registry=server_registry)
    else:
        server_registry = getattr(run_context.context, 'mcp_server_registry', None)
        if not server_registry:
            raise RuntimeError(
                "No server registry found in run context. Either specify it or set in context."
            )
        context = Context(server_registry=server_registry)

    # Create the aggregator
    aggregator = MCPAggregator(
        server_names=servers,
        connection_persistence=connection_persistence,
        name=name,
        context=context
    )

    return aggregator

async def initialize_mcp_aggregator(
    run_context: RunContextWrapper[TContext],
    name: str,
    servers: list[str],
    server_registry: ServerRegistry | None = None,
    connection_persistence: bool = True) -> MCPAggregator:
    """Initialize the MCP aggregator, which initializes all the server connections."""
    # Create the aggregator
    aggregator = create_mcp_aggregator(
        run_context=run_context,
        name=name,
        servers=servers,
        server_registry=server_registry,
        connection_persistence=connection_persistence
    )

    # Initialize the aggregator
    try:
        logger.info("Initializing MCPAggregator for %s with servers %s.", name, servers)
        await aggregator.__aenter__()
        logger.debug("MCPAggregator created and initialized for %s.", name)
        return aggregator
    except Exception as e:
        logger.error("Error creating MCPAggregator: %s", e)
        await aggregator.__aexit__(None, None, None)
        raise

def mcp_content_to_text(content: MCPContent | list[MCPContent]) -> str:
    """
    Convert CallToolResult MCP content to text.

    Args:
        content: MCP content object(s)

    Returns:
        String representation of the content
    """
    # Handle list of content items
    if isinstance(content, list):
        text_parts = []
        for item in content:
            if hasattr(item, 'type') and item.type == 'text' and hasattr(item, 'text'):
                # Text content
                text_parts.append(item.text)
            elif hasattr(item, 'type') and item.type == 'image' and hasattr(item, 'data'):
                # Image content - convert to text description
                mime_type = getattr(item, 'mimeType', 'unknown type')
                text_parts.append(f"[Image: {mime_type}]")
            elif hasattr(item, 'resource'):
                # Embedded resource
                resource = item.resource
                if hasattr(resource, 'text'):
                    text_parts.append(resource.text)
                elif hasattr(resource, 'blob'):
                    mime_type = getattr(resource, 'mimeType', 'unknown type')
                    text_parts.append(f"[Resource: {mime_type}]")
            else:
                # Unknown content type
                text_parts.append(str(item))

        if text_parts:
            return "\n".join(text_parts)
        return ""

    # Single content item
    if hasattr(content, 'type') and content.type == 'text' and hasattr(content, 'text'):
        return content.text
    elif hasattr(content, 'type') and content.type == 'image' and hasattr(content, 'data'):
        mime_type = getattr(content, 'mimeType', 'unknown type')
        return f"[Image: {mime_type}]"
    elif hasattr(content, 'resource'):
        resource = content.resource
        if hasattr(resource, 'text'):
            return resource.text
        elif hasattr(resource, 'blob'):
            mime_type = getattr(resource, 'mimeType', 'unknown type')
            return f"[Resource: {mime_type}]"

    # Fallback to string representation
    return str(content)

async def mcp_list_tools(server_aggregator: MCPAggregator) -> list[Tool]:
    """
    List all available tools from MCP servers that are part of the provided server aggregator.

    Args:
        server_aggregator: MCP server aggregator instance (must be initialized already)

    Returns:
        List of available tools
    """

    if not server_aggregator or server_aggregator.initialized is False:
        raise RuntimeError("MCP server aggregator not initialized when calling list_tools")

    # Get tools list from the aggregator
    tools_result = await server_aggregator.list_tools()

    # Convert MCP tools to OpenAI Agent SDK tools
    mcp_tools: list[Tool] = []
    for mcp_tool in tools_result.tools:
        tool = mcp_tool_to_function_tool(mcp_tool, server_aggregator)
        mcp_tools.append(tool)

    return mcp_tools


# JSON Schema properties not supported by OpenAI functions
UNSUPPORTED_SCHEMA_PROPERTIES = {
    "minimum", "minLength", "maxLength", "pattern", "format", "minItems", "maxItems", 
    "uniqueItems", "minProperties", "maxProperties", "multipleOf", 
    "exclusiveMinimum", "exclusiveMaximum", "$schema", "examples", "default"
}

def sanitize_json_schema_for_openai(schema: Dict[str, Any]) -> Dict[str, Any]:
    """
    Sanitize a JSON Schema to make it compatible with OpenAI function calling.
    Removes properties not supported by OpenAI's function schema validation.

    Args:
        schema: The original JSON schema

    Returns:
        A sanitized schema compatible with OpenAI
    """
    if not isinstance(schema, dict):
        return schema

    result = {}

    # Process each key in the schema
    for key, value in schema.items():
        # Skip unsupported properties
        if key in UNSUPPORTED_SCHEMA_PROPERTIES:
            continue

        # Handle nested objects recursively
        if isinstance(value, dict):
            result[key] = sanitize_json_schema_for_openai(value)
        # Handle arrays of objects
        elif isinstance(value, list):
            result[key] = [
                sanitize_json_schema_for_openai(item) if isinstance(item, dict) else item
                for item in value
            ]
        else:
            result[key] = value
    
    # Special handling for the properties/required issue
    # OpenAI requires all properties to be in the required array
    if "type" in result and result["type"] == "object" and "properties" in result:
        # Get all property names
        property_names = list(result.get("properties", {}).keys())
        
        # Set required field to include all properties
        if property_names:
            result["required"] = property_names

    return result

def mcp_tool_to_function_tool(
    mcp_tool: MCPTool, server_aggregator: MCPAggregator
) -> FunctionTool:
    """
    Convert an MCP tool to an OpenAI Agent SDK function tool.
    """

    # Create a properly named wrapper function
    function_name = _utils.transform_string_function_style(mcp_tool.name)

    # Create a wrapper factory to ensure each tool gets its own closure
    def create_wrapper(current_tool_name: str, current_tool_desc: str):
        async def wrapper_fn(ctx: RunContextWrapper[TContext], **kwargs: Any) -> Any:
            """MCP Tool wrapper function."""
            if not server_aggregator or server_aggregator.initialized is False:
                raise RuntimeError(
                    f"MCP aggregator not initialized for agent {server_aggregator.agent_name}"
                )

            # Call the tool through the aggregator
            result = await server_aggregator.call_tool(
                name=current_tool_name,
                arguments=kwargs
            )

            # Handle errors
            if getattr(result, 'isError', False):
                error_message = "Unknown error"
                # Try to extract error from content if available
                if hasattr(result, 'content'):
                    error_message = mcp_content_to_text(result.content)
                raise RuntimeError(
                    f"Error calling MCP tool '{current_tool_name}': {error_message}"
                )

            # Convert MCP content to string using helper method
            if hasattr(result, 'content'):
                return mcp_content_to_text(result.content)

            # Fallback for unexpected formats
            return str(result)

        # Set proper name and docstring for the function
        wrapper_fn.__name__ = function_name
        wrapper_fn.__doc__ = current_tool_desc or f"MCP tool: {current_tool_name}"
        return wrapper_fn

    # Create a wrapper for this specific tool
    tool_desc = mcp_tool.description or f"MCP tool: {mcp_tool.name}"
    wrapper_fn = create_wrapper(mcp_tool.name, tool_desc)

    # Create JSON schema for parameters - MCP uses inputSchema 
    params_schema = getattr(mcp_tool, 'inputSchema', {
        "type": "object",
        "properties": {},
        "required": [],
    })

    # OpenAI requires additionalProperties to be false for tool schemas
    params_schema["additionalProperties"] = False


    # Sanitize schema to remove properties not supported by OpenAI
    # OpenAI doesn't support minLength, maxLength, pattern, format, etc.
    params_schema = sanitize_json_schema_for_openai(params_schema)

    # Create a invoke tool factory to ensure each tool gets its own closure
    def create_invoke_tool(current_tool_name: str, current_wrapper_fn):
        async def invoke_tool(run_context: RunContextWrapper[Any], arguments_json: str) -> str:
            try:
                # Parse arguments from JSON
                import json
                args = json.loads(arguments_json)

                # Call the wrapper function with the arguments
                result = await current_wrapper_fn(run_context, **args)

                # Since wrapper_fn now returns string content (via _mcp_content_to_text),
                # we should already have a properly formatted string result
                return result
            except Exception as e:
                # Log the error
                logger.error(f"Error invoking MCP tool {current_tool_name}: {e}")

                # Format error message
                error_type = type(e).__name__
                error_message = str(e)

                # Return error message that's helpful to the model
                return f"Error ({error_type}): {error_message}"
        return invoke_tool

    # Create the invoke tool function specific to this tool
    invoke_tool = create_invoke_tool(mcp_tool.name, wrapper_fn)

    # Create a function tool
    tool = FunctionTool(
        name=mcp_tool.name,
        description=mcp_tool.description or f"MCP tool: {mcp_tool.name}",
        params_json_schema=params_schema,
        on_invoke_tool=invoke_tool,
    )

    return tool



from __future__ import annotations

import dataclasses
import inspect
from collections.abc import Awaitable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Generic, cast

from . import _utils
from ._utils import MaybeAwaitable
from .guardrail import InputGuardrail, OutputGuardrail
from .handoffs import Handoff
from .items import ItemHelpers
from .logger import logger
from .model_settings import ModelSettings
from .models.interface import Model
from .run_context import RunContextWrapper, TContext
from .tool import Tool, function_tool

if TYPE_CHECKING:
    from mcp_agent.mcp.mcp_aggregator import MCPAggregator
    from mcp_agent.mcp_server_registry import ServerRegistry

    from .lifecycle import AgentHooks
    from .result import RunResult


@dataclass
class Agent(Generic[TContext]):
    """An agent is an AI model configured with instructions, tools, guardrails, handoffs and more.

    We strongly recommend passing `instructions`, which is the "system prompt" for the agent. In
    addition, you can pass `description`, which is a human-readable description of the agent, used
    when the agent is used inside tools/handoffs.

    Agents are generic on the context type. The context is a (mutable) object you create. It is
    passed to tool functions, handoffs, guardrails, etc.
    """

    name: str
    """The name of the agent."""

    instructions: (
        str
        | Callable[
            [RunContextWrapper[TContext], Agent[TContext]],
            MaybeAwaitable[str],
        ]
        | None
    ) = None
    """The instructions for the agent. Will be used as the "system prompt" when this agent is
    invoked. Describes what the agent should do, and how it responds.

    Can either be a string, or a function that dynamically generates instructions for the agent. If
    you provide a function, it will be called with the context and the agent instance. It must
    return a string.
    """

    handoff_description: str | None = None
    """A description of the agent. This is used when the agent is used as a handoff, so that an
    LLM knows what it does and when to invoke it.
    """

    handoffs: list[Agent[Any] | Handoff[TContext]] = field(default_factory=list)
    """Handoffs are sub-agents that the agent can delegate to. You can provide a list of handoffs,
    and the agent can choose to delegate to them if relevant. Allows for separation of concerns and
    modularity.
    """

    model: str | Model | None = None
    """The model implementation to use when invoking the LLM.

    By default, if not set, the agent will use the default model configured in
    `model_settings.DEFAULT_MODEL`.
    """

    model_settings: ModelSettings = field(default_factory=ModelSettings)
    """Configures model-specific tuning parameters (e.g. temperature, top_p).
    """

    tools: list[Tool] = field(default_factory=list)
    """A list of tools that the agent can use."""

    input_guardrails: list[InputGuardrail[TContext]] = field(default_factory=list)
    """A list of checks that run in parallel to the agent's execution, before generating a
    response. Runs only if the agent is the first agent in the chain.
    """

    output_guardrails: list[OutputGuardrail[TContext]] = field(default_factory=list)
    """A list of checks that run on the final output of the agent, after generating a response.
    Runs only if the agent produces a final output.
    """

    output_type: type[Any] | None = None
    """The type of the output object. If not provided, the output will be `str`."""

    hooks: AgentHooks[TContext] | None = None
    """A class that receives callbacks on various lifecycle events for this agent.
    """

    mcp_servers: list[str] = field(default_factory=list)
    """A list of MCP server names to use with this agent.

    The agent will automatically discover and include tools from these MCP servers.
    Each server name must be registered in the server registry,
    which is initialized from mcp_agent.config.yaml.
    """

    mcp_server_registry: ServerRegistry | None = None
    """The server registry to use with this agent. 
    If not provided, it will be loaded from the run context, which initializes it from mcp_agent.config.yaml.
    """

    _mcp_aggregator: MCPAggregator | None = None
    """The MCP aggregator used by this agent. Will be created lazily when needed."""

    _mcp_initialized: bool = False
    """Whether MCP tools have been loaded for this agent."""

    def clone(self, **kwargs: Any) -> Agent[TContext]:
        """Make a copy of the agent, with the given arguments changed. For example, you could do:
        ```
        new_agent = agent.clone(instructions="New instructions")
        ```
        """
        return dataclasses.replace(self, **kwargs)

    def as_tool(
        self,
        tool_name: str | None,
        tool_description: str | None,
        custom_output_extractor: Callable[[RunResult], Awaitable[str]] | None = None,
    ) -> Tool:
        """Transform this agent into a tool, callable by other agents.

        This is different from handoffs in two ways:
        1. In handoffs, the new agent receives the conversation history. In this tool, the new agent
           receives generated input.
        2. In handoffs, the new agent takes over the conversation. In this tool, the new agent is
           called as a tool, and the conversation is continued by the original agent.

        Args:
            tool_name: The name of the tool. If not provided, the agent's name will be used.
            tool_description: The description of the tool, which should indicate what it does and
                when to use it.
            custom_output_extractor: A function that extracts the output from the agent. If not
                provided, the last message from the agent will be used.
        """

        @function_tool(
            name_override=tool_name or _utils.transform_string_function_style(self.name),
            description_override=tool_description or "",
        )
        async def run_agent(context: RunContextWrapper, input: str) -> str:
            from .run import Runner

            output = await Runner.run(
                starting_agent=self,
                input=input,
                context=context.context,
            )
            if custom_output_extractor:
                return await custom_output_extractor(output)

            return ItemHelpers.text_message_outputs(output.new_items)

        return run_agent

    async def get_system_prompt(self, run_context: RunContextWrapper[TContext]) -> str | None:
        """Get the system prompt for the agent."""
        if isinstance(self.instructions, str):
            return self.instructions
        elif callable(self.instructions):
            if inspect.iscoroutinefunction(self.instructions):
                return await cast(Awaitable[str], self.instructions(run_context, self))
            else:
                return cast(str, self.instructions(run_context, self))
        elif self.instructions is not None:
            logger.error(f"Instructions must be a string or a function, got {self.instructions}")

        return None

    async def load_mcp_tools(self, run_context: RunContextWrapper[TContext]) -> None:
        """Load tools from MCP servers and add them to this agent's tools."""

        logger.debug(f"MCP servers: {self.mcp_servers}")

        if not self.mcp_servers or self._mcp_initialized:
            return

        from .mcp import initialize_mcp_aggregator, mcp_list_tools

        if self._mcp_aggregator is None:
            self._mcp_aggregator = await initialize_mcp_aggregator(
                run_context,
                name=self.name,
                servers=self.mcp_servers,
                server_registry=self.mcp_server_registry,
                connection_persistence=True
            )

        # Get all tools from the MCP servers
        mcp_tools = await mcp_list_tools(self._mcp_aggregator)

        # Add the MCP tools to the agent's tools
        logger.info(f"Adding {len(mcp_tools)} MCP tools to agent {self.name}")
        self.tools.extend(mcp_tools)
        self._mcp_initialized = True

    async def cleanup_resources(self) -> None:
        """Clean up resources when the agent is done."""
        if self._mcp_aggregator:
            logger.info(f"Cleaning up MCP resources for agent {self.name}")
            try:
                await self._mcp_aggregator.__aexit__(None, None, None)
                self._mcp_aggregator = None
                self._mcp_initialized = False
            except Exception as e:
                logger.error(f"Error cleaning up MCP resources for agent {self.name}: {e}")
